/* combustion of methane with dissociation
   CH4+2O2=>C+CO+CO2+H+H2+H2O+O+O2+OH
   using SI units
   developed by Dudley J. Benton */

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stdlib.h>
#include <malloc.h>
#include <float.h>
#define _USE_MATH_DEFINES
#include <math.h>

#define Pref 0.101325  /* MPa */
#define Tref 298.15    /* K */
#define R    8.3144621 /* kJ/kg-mole/K */

typedef struct{double MW,Hf,Sf,Cp;}PROP; /* kJ/kg/K */

PROP C  ={12.011, 717162.,158.1,20.85}; /* free carbon */
PROP CH4={16.043, -74902.,255.2,20.55}; /* methane */
PROP CO ={28.010,-110615.,198.0,28.10}; /* carbon monoxide */
PROP CO2={44.009,-393769.,213.8,33.57}; /* carbon dioxide */
PROP H  ={ 1.008, 218111.,114.7,20.80}; /* monatomic hydrogen */
PROP H2 ={ 2.016,      0.,130.7,27.33}; /* diatomic hydrogen */
PROP H2O={18.015,-241981.,188.8,28.77}; /* water (vapor) */
PROP O  ={15.999, 249336.,161.1,21.93}; /* monatomic oxygen  */
PROP O2 ={31.998,      0.,205.2,28.01}; /* diatomic oxygen */
PROP OH ={17.007,  39733.,183.7,28.29}; /* hydroxl */

char*Name[]={"C","CH4","CO","CO2","H","H2","H2O","O","O2","OH"};

double Q=0.;    /* heat transfer into the system (assumed adiabatic) */
double P1=Pref; /* initial pressure */
double P2=Pref; /* final pressure (assumed isobaric) */
double T1=Tref; /* initial temperature */
double T2;      /* final temperature (unknown) */

#define nE 3 /* number of elements (C, H, and O) */
#define nR 2 /* number of reactants (CH4 and O2) */
#define nP 9 /* number of products (C, CO, CO2, H, H2, H2O, O, O2, and OH) */

PROP pR[nR]; /* reactant properties */
PROP pP[nP]; /* product properties */

double yR[nR]={1.,2.}; /* reactant moles */
double yP[nP];         /* product moles (unknown) */
double yX[nP];         /* maximum possible moles of each product */

double A[(nP+nE)*(nP+nE)]; /* combined matrix [ACTC0] */
double B[nP+nE];           /* combined matrix [BD] */
double x[nP+nE];           /* combined matrix [yL] */
double c[nE*(nP+nE)];      /* product elemental abundances */
double D[nE];              /* reactant elemental abundances */

double fmin(double x,double y)
  {
  if(x<y)
    return(x);
  return(y);
  }

double fmax(double x,double y)
  {
  if(x>y)
    return(x);
  return(y);
  }

int GaussElimination(double*A,double*B,int n)
  {
  int i,j,k,ip,jp,*pivot;
  double a,p,b;
  if(n<1)
    return(__LINE__);
  if(n==1)
    {
    a=A[0];
    if(fabs(a)<FLT_MIN)
      return(__LINE__);
    B[0]/=a;
    return(0);
    }
  if((pivot=calloc(n,sizeof(int)))==NULL)
    return(__LINE__);
  for(i=0;i<n;i++)
    pivot[i]=i;
  for(k=0;k<n-1;k++)
    {
    ip=k;
    jp=k;
    p=fabs(A[k*(n+1)]);
    for(i=k;i<n;i++)
      {
      for(j=k;j<n;j++)
        {
        a=fabs(A[n*i+j]);
        if(a>p)
          {
          ip=i;
          jp=j;
          p=a;
          }
        }
      }
    if(p<FLT_MIN)
      {
      free(pivot);
      return(__LINE__);
      }
    if(ip!=k)
      {
      b=B[k];
      B[k]=B[ip];
      B[ip]=b;
      for(j=k;j<n;j++)
        {
        a=A[n*k+j];
        A[n*k+j]=A[n*ip+j];
        A[n*ip+j]=a;
        }
      }
    if(jp!=k)
      {
      j=pivot[jp];
      pivot[jp]=pivot[k];
      pivot[k]=j;
      for(i=0;i<n;i++)
        {
        a=A[n*i+k];
        A[n*i+k]=A[n*i+jp];
        A[n*i+jp]=a;
        }
      }
    for(i=k+1;i<n;i++)
      {
      a=A[n*i+k]/A[n*k+k];
      B[i]-=a*B[k];
      for(j=k+1;j<n;j++)
        A[n*i+j]-=a*A[n*k+j];
      }
    }
  a=A[n*n-1];
  p=fabs(a);
  if(p<FLT_MIN)
    {
    free(pivot);
    return(__LINE__);
    }
  B[n-1]/=a;
  for(k=1;k<n;k++)
    {
    i=n-1-k;
    b=0.;
    for(j=i+1;j<n;j++)
      b+=A[n*i+j]*B[j];
    B[i]=(B[i]-b)/A[n*i+i];
    }
  for(i=0;i<n;i++)
    A[i]=B[i];
  for(i=0;i<n;i++)
    B[pivot[i]]=A[i];
  free(pivot);
  return(0);
  }

double h(PROP prop,double T) /* enthalpy of one product */
  {
  return(prop.Hf+prop.Cp*(T-Tref));
  }

double hR() /* enthalpy of all reactants combined */
  {
  int i;
  double hr;
  for(hr=i=0;i<nR;i++)
    hr+=yR[i]*h(pR[i],T1);
  return(hr);
  }

double hP() /* enthalpy of all products combined */
  {
  int i;
  double hp;
  for(hp=i=0;i<nP;i++)
    hp+=yP[i]*h(pP[i],T2);
  return(hp);
  }

double g(PROP prop,double T,double P) /* Gibbs free energy of one product */
  {
  return(prop.Hf+prop.Cp*(T-Tref)-T*(prop.Sf+prop.Cp*log(T/Tref)-R*log(P/Pref)));
  }

double G() /* Gibbs free energy of all products combined */
  {
  int i;
  double G,yS;
  for(yS=i=0;i<nP;i++)
    yS+=yP[i];
  for(G=i=0;i<nP;i++)
    G+=yP[i]*g(pP[i],T2,yP[i]*P2/Pref/yS);
  return(G);
  }

int main(int argc,char**argv,char**envp)
  {
  int i,iter,j,k,n=nP+nE;
  double damp=2.,yS;

/* copy properties into arrays for convenience loops */
/* (these don't change with iterations) */

  pR[0]=CH4;
  pR[1]=O2;
  pP[0]=C;
  pP[1]=CO;
  pP[2]=CO2;
  pP[3]=H;
  pP[4]=H2;
  pP[5]=H2O;
  pP[6]=O;
  pP[7]=O2;
  pP[8]=OH;

/* carbon abundance constraints */
/* (these don't change with iterations) */

  c[n*0+0]=1.; /* C */
  c[n*0+1]=1.; /* CO */
  c[n*0+2]=1.; /* CO2 */
  c[n*0+3]=0.; /* H */
  c[n*0+4]=0.; /* H2 */
  c[n*0+5]=0.; /* H2O */
  c[n*0+6]=0.; /* O */
  c[n*0+7]=0.; /* O2 */
  c[n*0+8]=0.; /* OH */
  D[0]=1.;     /* CH4 */

/* hydrogen abundance constraints */
/* (these don't change with iterations) */

  c[n*1+0]=0.; /* C */
  c[n*1+1]=0.; /* CO */
  c[n*1+2]=0.; /* CO2 */
  c[n*1+3]=1.; /* H */
  c[n*1+4]=2.; /* H2 */
  c[n*1+5]=2.; /* H2O */
  c[n*1+6]=0.; /* O */
  c[n*1+7]=0.; /* O2 */
  c[n*1+8]=1.; /* OH */
  D[1]=4.;     /* CH4 */

/* oxygen abundance constraints */
/* (these don't change with iterations) */

  c[n*2+0]=0.; /* C */
  c[n*2+1]=1.; /* CO */
  c[n*2+2]=2.; /* CO2 */
  c[n*2+3]=0.; /* H */
  c[n*2+4]=0.; /* H2 */
  c[n*2+5]=1.; /* H2O */
  c[n*2+6]=1.; /* O */
  c[n*2+7]=2.; /* O2 */
  c[n*2+8]=1.; /* OH */
  D[2]=4.;     /* 2O2 */

/* maximum possible product moles */

  yX[0]=1.; /* C */
  yX[1]=1.; /* CO */
  yX[2]=1.; /* CO2 */
  yX[3]=4.; /* H */
  yX[4]=2.; /* H2 */
  yX[5]=2.; /* H2O */
  yX[6]=4.; /* O */
  yX[7]=2.; /* O2 */
  yX[8]=4.; /* OH */

/* initial estimate of final temperature */
/* (this will be corrected after each iteration) */

  T2=3400.;

/* initial estimate of product moles */
/* (necessary to initialize iterations) */

  for(i=0;i<nP;i++)
    yP[i]=yX[i]/4.;

/* fill in free energy minimization terms */
/* (these change with each iteration) */

  for(iter=0;iter<32;iter++)
    {
    printf("iter=%i, ",iter);
    memset(A,0,sizeof(A)); /* zero out matix A (zero B not necessary) */
    for(yS=i=0;i<nP;i++)
      yS+=yP[i];
    for(i=0;i<nP;i++)
      {
      for(j=0;j<nP;j++)
        {
        if(j==i)
          A[n*i+i]=R*T2*(yS-yP[i])/yP[i]/yS;
        else
          A[n*i+j]=-R*T2/yS;
        }
      B[i]=-g(pP[i],T2,yP[i]*P2/Pref/yS);
      }

/* add C, C-transpose, and D */
/* (these don't change, but the matrices A and B are destroyed) */

    for(i=0;i<nE;i++)
      {
      k=nP+i;
      for(j=0;j<nP;j++)
        A[n*k+j]=A[n*j+k]=c[n*i+j];
      B[k]=D[i];
      }

/* solve simultaneous linear equations */

    i=GaussElimination(A,B,n);
    printf("error=%i\n",i);
    for(i=0;i<nP;i++)
      {
      yP[i]=fmax(yP[i]/2.,fmin((yP[i]+yX[i])/2.,(damp*yP[i]+B[i])/(damp+1.)));
      printf("%-3s=%.12lf\n",Name[i],yP[i]);
      }
    printf("T2=%.0lf, hR=%.0lf, hP=%.0lf\n",T2,hR(),hP());
    T2+=(hR()+Q-hP())/R/50.;
    }

  return(0);
  }
